@@ -1,24 +1,36 @@
-package # not an official package
-OpenGL::Modern::Helpers;
+package    # not an official package
+  OpenGL::Modern::Helpers;
 
 use strict;
 use Exporter 'import';
 use Carp qw(croak);
+use Config;
 
 use OpenGL::Modern qw(
-    GL_NO_ERROR
-    GL_INVALID_ENUM
-    GL_INVALID_VALUE
-    GL_INVALID_OPERATION
-    GL_STACK_OVERFLOW
-    GL_STACK_UNDERFLOW
-    GL_OUT_OF_MEMORY
-    GL_TABLE_TOO_LARGE
-    GL_VERSION
-    glGetString
-    glGetError
-    glGetShaderInfoLog_c
-    glGetProgramInfoLog_c
+  GL_NO_ERROR
+  GL_INVALID_ENUM
+  GL_INVALID_VALUE
+  GL_INVALID_OPERATION
+  GL_STACK_OVERFLOW
+  GL_STACK_UNDERFLOW
+  GL_OUT_OF_MEMORY
+  GL_TABLE_TOO_LARGE
+  GL_VERSION
+  glGetString
+  glGetError
+  glGetShaderInfoLog_c
+  glGetProgramInfoLog_c
+  glGenTextures_c
+  glGetProgramiv_c
+  glGetShaderiv_c
+  glShaderSource_c
+  glGenFramebuffers_c
+  glGenVertexArrays_c
+  glGenBuffers_c
+  glGetIntegerv_c
+  glBufferData_c
+  glUniform2f
+  glUniform4f
 );
 
 =head1 NAME
@@ -131,57 +143,68 @@ use vars qw(@EXPORT_OK $VERSION %glErrorStrings);
 $VERSION = '0.01_02';
 
 @EXPORT_OK = qw(
-    pack_GLuint
-    pack_GLfloat
-    pack_GLdouble
-    pack_GLint
-    pack_GLstrings
-    pack_ptr
-    xs_buffer
-    
-    glGetShaderInfoLog_p
-    glGetProgramInfoLog_p
-    croak_on_gl_error
-        
-    glGetVersion_p
-);
+  pack_GLuint
+  pack_GLfloat
+  pack_GLdouble
+  pack_GLint
+  pack_GLstrings
+  pack_ptr
+  xs_buffer
 
+  glGetShaderInfoLog_p
+  glGetProgramInfoLog_p
+  croak_on_gl_error
+
+  glGetVersion_p
+  glGenTextures_p
+  glGetProgramiv_p
+  glGetShaderiv_p
+  glShaderSource_p
+  glGenFramebuffers_p
+  glGenVertexArrays_p
+  glGenBuffers_p
+  glGetIntegerv_p
+  glBufferData_p
+  glUniform2f_p
+  glUniform4f_p
+);
 
 %glErrorStrings = (
-    GL_NO_ERROR() => 'No error has been recorded.',
-    GL_INVALID_ENUM() => 'An unacceptable value is specified for an enumerated argument.',
-    GL_INVALID_VALUE() => 'A numeric argument is out of range.',
+    GL_NO_ERROR()          => 'No error has been recorded.',
+    GL_INVALID_ENUM()      => 'An unacceptable value is specified for an enumerated argument.',
+    GL_INVALID_VALUE()     => 'A numeric argument is out of range.',
     GL_INVALID_OPERATION() => 'The specified operation is not allowed in the current state.',
-    GL_STACK_OVERFLOW() => 'This command would cause a stack overflow.',
-    GL_STACK_UNDERFLOW() => 'This command would cause a stack underflow.',
-    GL_OUT_OF_MEMORY() => 'There is not enough memory left to execute the command.',
-    GL_TABLE_TOO_LARGE() => 'The specified table exceeds the implementation\'s maximum supported table size.',
+    GL_STACK_OVERFLOW()    => 'This command would cause a stack overflow.',
+    GL_STACK_UNDERFLOW()   => 'This command would cause a stack underflow.',
+    GL_OUT_OF_MEMORY()     => 'There is not enough memory left to execute the command.',
+    GL_TABLE_TOO_LARGE()   => 'The specified table exceeds the implementation\'s maximum supported table size.',
 );
 
+our $PACK_TYPE = $Config{ptrsize} == 4 ? 'L' : 'Q';
 
 sub pack_GLuint {
     my @gluints = @_;
-    pack 'I*', @gluints
+    pack 'I*', @gluints;
 }
 
 sub pack_GLint {
     my @gluints = @_;
-    pack 'I*', @gluints
+    pack 'I*', @gluints;
 }
 
 sub pack_GLfloat {
     my @glfloats = @_;
-    pack 'f*', @glfloats
+    pack 'f*', @glfloats;
 }
 
 sub pack_GLdouble {
     my @gldoubles = @_;
-    pack 'd*', @gldoubles
+    pack 'd*', @gldoubles;
 }
 
 # No parameter declaration because we don't want copies
 sub pack_GLstrings {
-    pack 'P*', @_
+    pack 'P*', @_;
 }
 
 # No parameter declaration because we don't want copies
@@ -205,41 +228,98 @@ sub xs_buffer {
     $_[0];
 }
 
-sub glGetShaderInfoLog_p {
-    my $shader = $_[0];
-    my $bufsize = 1024*64;
-    my $buffer = "\0" x $bufsize;
-    my $len = "\0" x 4;
+sub get_info_log_p {
+    my ( $call, $id ) = @_;
+    my $bufsize = 1024 * 64;
+    my $buffer  = "\0" x $bufsize;
+    my $len     = "\0" x 4;
+
     # void glGetShaderInfoLog(GLuint shader, GLsizei bufSize, GLsizei* length, GLchar* infoLog);
-    glGetShaderInfoLog_c( $shader, $bufsize, unpack('Q',pack('p',$len)), $buffer);
+    # void glGetProgramInfoLog(GLuint program, GLsizei bufSize, GLsizei* length, GLchar* infoLog);
+    $call->( $id, $bufsize, unpack( $PACK_TYPE, pack( 'p', $len ) ), $buffer );
     $len = unpack 'I', $len;
     return substr $buffer, 0, $len;
 }
 
-sub glGetProgramInfoLog_p {
-    my $program = $_[0];
-    my $bufsize = 1024*64;
-    my $buffer = "\0" x $bufsize;
-    my $len = "\0" x 4;
-    # void glGetProgramInfoLog(GLuint program, GLsizei bufSize, GLsizei* length, GLchar* infoLog);
-    glGetProgramInfoLog_c( $program, $bufsize, unpack('Q',pack('p',$len)), $buffer);
-    $len = unpack 'I', $len;
-    return substr $buffer, 0, $len;
-}
+sub glGetShaderInfoLog_p  { get_info_log_p \&glGetShaderInfoLog,  @_ }
+sub glGetProgramInfoLog_p { get_info_log_p \&glGetProgramInfoLog, @_ }
 
 sub glGetVersion_p {
+
     # const GLubyte * GLAPIENTRY glGetString (GLenum name);
-    my $glVersion = glGetString(GL_VERSION);
-    ($glVersion) = ($glVersion =~ m!^(\d+\.\d+)!g);
-    $glVersion
+    my $glVersion = glGetString( GL_VERSION );
+    ( $glVersion ) = ( $glVersion =~ m!^(\d+\.\d+)!g );
+    $glVersion;
 }
 
 sub croak_on_gl_error {
+
     # GLenum glGetError (void);
     my $error = glGetError();
-    if( $error != GL_NO_ERROR ) {
-        croak $glErrorStrings{ $error } || "Unknown OpenGL error: $error"
-    };
+    if ( $error != GL_NO_ERROR ) {
+        croak $glErrorStrings{$error} || "Unknown OpenGL error: $error";
+    }
+}
+
+sub gen_thing_p {
+    my ( $call, $n ) = @_;
+    xs_buffer my $new_ids, 4 * $n;
+    $call->( $n, unpack( $PACK_TYPE, pack( 'p', $new_ids ) ) );
+    my @ids = unpack 'I*', $new_ids;
+    return wantarray ? @ids : $ids[0];
+}
+
+sub glGenTextures_p { gen_thing_p \&glGenTextures_c, @_ }
+
+sub glGenFramebuffers_p { gen_thing_p \&glGenFramebuffers_c, @_ }
+
+sub glGenVertexArrays_p { gen_thing_p \&glGenVertexArrays_c, @_ }
+
+sub glGenBuffers_p { gen_thing_p \&glGenBuffers_c, @_ }
+
+sub get_iv_p {
+    my ( $call, $id, $pname, $count ) = @_;
+    $count ||= 1;
+    xs_buffer my $params, 4 * $count;
+    $call->( $id, $pname, unpack( "$PACK_TYPE*", pack( 'p*', $params ) ) );
+    my @params = unpack 'I*', $params;
+    return wantarray ? @params : $params[0];
+}
+
+sub glGetProgramiv_p { get_iv_p \&glGetProgramiv_c, @_ }
+
+sub glGetShaderiv_p { get_iv_p \&glGetShaderiv_c, @_ }
+
+sub glShaderSource_p {
+    my ( $shader, @sources ) = @_;
+    my $count = @sources;
+    my @lengths = map length, @sources;
+    glShaderSource_c( $shader, $count, pack( 'P*', @sources ), pack( 'I*', @lengths ) );
+    return;
+}
+
+sub glGetIntegerv_p {
+    my ( $pname, $count ) = @_;
+    $count ||= 1;
+    xs_buffer my $data, 4 * $count;
+    glGetIntegerv_c $pname, unpack( $PACK_TYPE, pack( 'p', $data ) );
+    my @data = unpack 'I*', $data;
+    return wantarray ? @data : $data[0];
+}
+
+sub glBufferData_p {
+    my ( $target, $oga, $usage ) = @_;
+    glBufferData_c $target, $oga->length, $oga->ptr, $usage;
+}
+
+sub glUniform2f_p {
+    my ( $uniform, $v0, $v1 ) = @_;
+    glUniform2f $uniform, $v0, $v1;
+}
+
+sub glUniform4f_p {
+    my ( $uniform, $v0, $v1, $v2, $v3 ) = @_;
+    glUniform4f $uniform, $v0, $v1, $v2, $v3;
 }
 
 1;
